@@ -2,7 +2,7 @@
 //!
 //! Objects taken directly from C are typically prefixed with "Image" and will closely
 //! resemble the names of their C counterparts, but named to conform to Rust standards.
-//! For example, ```IMAGE_DIRECTORY_ENTRY``` is known as [ImageDirectoryEntry](ImageDirectoryEntry) in
+//! For example, ```IMAGE_DIRECTORY_ENTRY``` is known as [`ImageDirectoryEntry`](ImageDirectoryEntry) in
 //! this library.
 
 use bitflags::bitflags;
@@ -18,7 +18,7 @@ use std::default::Default;
 use std::mem;
 use std::slice;
 
-use crate::{PE, PEType, PETranslation, Error};
+use crate::*;
 use crate::types::*;
 
 pub const DOS_SIGNATURE: u16    = 0x5A4D;
@@ -753,6 +753,8 @@ impl ImageSectionHeader {
     }
 
     /// Read a slice of the data this section represents.
+    ///
+    /// The address and size chosen is relative to the PE argument's [type](PEType).
     pub fn read<'data>(&'data self, pe: &'data PE) -> Result<&'data [u8], Error> {
         let offset = self.data_offset(pe.pe_type);
         let size = self.data_size(pe.pe_type);
@@ -760,6 +762,8 @@ impl ImageSectionHeader {
         pe.buffer.read(offset, size)
     }
     /// Read a mutable slice of the data this section represents.
+    ///
+    /// The address and size chosen is relative to the PE argument's [type](PEType).
     pub fn read_mut<'data>(&'data self, pe: &'data mut PE) -> Result<&'data mut [u8], Error> {
         let offset = self.data_offset(pe.pe_type);
         let size = self.data_size(pe.pe_type);
@@ -768,6 +772,8 @@ impl ImageSectionHeader {
     }
     /// Write data to this section. It returns [`Error::BufferTooSmall`](Error::BufferTooSmall) if the data
     /// overflows the section.
+    ///
+    /// The address and size chosen is relative to the PE argument's [type](PEType).
     pub fn write(&self, pe: &mut PE, data: &[u8]) -> Result<(), Error> {
         let offset = self.data_offset(pe.pe_type);
         let size = self.data_size(pe.pe_type);
@@ -834,76 +840,6 @@ pub struct ImageDataDirectory {
     pub virtual_address: RVA,
     pub size: u32,
 }
-impl ImageDataDirectory {
-    /// Get the data directory object pointed to by this data directory entry.
-    pub fn resolve<'data>(&'data self, pe: &'data PE, entry: ImageDirectoryEntry) -> Result<DataDirectory<'data>, Error> {
-        if self.virtual_address.0 == 0 {
-            return Err(Error::InvalidRVA);
-        }
-        
-        let address = match pe.translate(PETranslation::Memory(self.virtual_address)) {
-            Ok(a) => a,
-            Err(e) => return Err(e),
-        };
-
-        /* we use an if/else statement instead of a match block for readability */
-        if entry == ImageDirectoryEntry::Export {
-            match pe.buffer.get_ref::<ImageExportDirectory>(address) {
-                Ok(d) => Ok(DataDirectory::Export(d)),
-                Err(e) => Err(e),
-            }
-        }
-        else if entry == ImageDirectoryEntry::Import {
-            match ImageImportDescriptor::parse_import_table(pe, self) {
-                Ok(d) => Ok(DataDirectory::Import(d)),
-                Err(e) => Err(e),
-            }
-        }
-        else if entry == ImageDirectoryEntry::BaseReloc {
-            match RelocationEntry::parse_table(pe, self) {
-                Ok(d) => Ok(DataDirectory::BaseReloc(d)),
-                Err(e) => Err(e),
-            }
-        }
-        else {
-            Err(Error::UnsupportedDirectory)
-        }
-    }
-    /// Get the mutable data directory object pointed to by this data directory entry.
-    pub fn resolve_mut<'data>(&'data self, pe: &'data mut PE, entry: ImageDirectoryEntry) -> Result<DataDirectoryMut<'data>, Error> {
-        if self.virtual_address.0 == 0 {
-            return Err(Error::InvalidRVA);
-        }
-        
-        let address = match pe.translate(PETranslation::Memory(self.virtual_address)) {
-            Ok(a) => a,
-            Err(e) => return Err(e),
-        };
-
-        /* we use an if/else statement instead of a match block for readability */
-        if entry == ImageDirectoryEntry::Export {
-            match pe.buffer.get_mut_ref::<ImageExportDirectory>(address) {
-                Ok(d) => Ok(DataDirectoryMut::Export(d)),
-                Err(e) => Err(e),
-            }
-        }
-        else if entry == ImageDirectoryEntry::Import {
-            match ImageImportDescriptor::parse_mut_import_table(pe, self) {
-                Ok(d) => Ok(DataDirectoryMut::Import(d)),
-                Err(e) => Err(e),
-            }
-        }
-        else if entry == ImageDirectoryEntry::BaseReloc {
-            match RelocationEntryMut::parse_table(pe, self) {
-                Ok(d) => Ok(DataDirectoryMut::BaseReloc(d)),
-                Err(e) => Err(e),
-            }
-        }
-        else {
-            Err(Error::UnsupportedDirectory)
-        }
-    }
-}
 impl Clone for ImageDataDirectory {
     fn clone(&self) -> Self {
         Self {
@@ -933,6 +869,44 @@ pub struct ImageExportDirectory {
     pub address_of_name_ordinals: RVA, // [u16; number_of_names]
 }
 impl ImageExportDirectory {
+    /// Parse the export table in the PE file.
+    pub fn parse<'data>(pe: &'data PE) -> Result<&'data ImageExportDirectory, Error> {
+        let dir = match pe.get_data_directory(ImageDirectoryEntry::Export) {
+            Ok(d) => d,
+            Err(e) => return Err(e),
+        };
+
+        if dir.virtual_address.0 == 0 || !pe.validate_rva(dir.virtual_address) {
+            return Err(Error::InvalidRVA);
+        }
+
+        let offset = match pe.translate(PETranslation::Memory(dir.virtual_address)) {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+
+        pe.buffer.get_ref::<ImageExportDirectory>(offset)
+    }
+
+    /// Parse a mutable export table in the PE file.
+    pub fn parse_mut<'data>(pe: &'data mut PE) -> Result<&'data mut ImageExportDirectory, Error> {
+        let dir = match pe.get_data_directory(ImageDirectoryEntry::Export) {
+            Ok(d) => d,
+            Err(e) => return Err(e),
+        };
+
+        if dir.virtual_address.0 == 0 || !pe.validate_rva(dir.virtual_address) {
+            return Err(Error::InvalidRVA);
+        }
+
+        let offset = match pe.translate(PETranslation::Memory(dir.virtual_address)) {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+
+        pe.buffer.get_mut_ref::<ImageExportDirectory>(offset)
+    }
+    
     /// Get the name of this export module.
     pub fn get_name<'data>(&self, pe: &'data PE) -> Result<&'data [CChar], Error> {
         if self.name.0 == 0 {
@@ -956,8 +930,8 @@ impl ImageExportDirectory {
         }
     }
     /// Get the function array of this export entry. This array represents thunk data pointing to either
-    /// ordinals [ThunkData::Ordinal](ThunkData::Ordinal), forwarder strings ([ThunkData::ForwarderString](ThunkData::ForwarderString)
-    /// or function data [ThunkData::Function](ThunkData::Function).
+    /// ordinals [`ThunkData::Ordinal`](ThunkData::Ordinal), forwarder strings ([`ThunkData::ForwarderString`](ThunkData::ForwarderString)
+    /// or function data [`ThunkData::Function`](ThunkData::Function).
     pub fn get_functions<'data>(&self, pe: &'data PE) -> Result<&'data [Thunk32], Error> {
         if self.address_of_functions.0 == 0 {
             return Err(Error::InvalidRVA);
@@ -1026,8 +1000,8 @@ impl ImageExportDirectory {
         }
     }
     /// Get a mapping of exports to thunk data for this export entry. This maps exported names to thunk data, which can
-    /// be an ordinal ([ThunkData::Ordinal](ThunkData::Ordinal)), a function ([ThunkData::Function](ThunkData::Function))
-    /// or a forwarder string ([ThunkData::ForwarderString](ThunkData::ForwarderString)).
+    /// be an ordinal ([`ThunkData::Ordinal`](ThunkData::Ordinal)), a function ([`ThunkData::Function`](ThunkData::Function))
+    /// or a forwarder string ([`ThunkData::ForwarderString`](ThunkData::ForwarderString)).
     pub fn get_export_map<'data>(&self, pe: &'data PE) -> Result<HashMap<&'data str, ThunkData>, Error> {
         let mut result: HashMap<&'data str, ThunkData> = HashMap::<&'data str, ThunkData>::new();
 
@@ -1221,68 +1195,6 @@ impl ImageImportDescriptor {
         }
     }
 
-    /// Parse the size of the import table pointed to by the ```ImageDataDirectory``` reference.
-    pub fn parse_import_table_size(pe: &PE, dir: &ImageDataDirectory) -> Result<usize, Error> {
-        if dir.virtual_address.0 == 0 {
-            return Err(Error::InvalidRVA);
-        }
-
-        let mut address = match pe.translate(PETranslation::Memory(dir.virtual_address)) {
-            Ok(a) => a,
-            Err(e) => return Err(e),
-        };
-
-        let mut imports = 0usize;
-
-        loop {
-            if !pe.validate_offset(address) {
-                return Err(Error::InvalidOffset);
-            }
-
-            match pe.buffer.get_ref::<ImageImportDescriptor>(address) {
-                Ok(x) => { if x.original_first_thunk.0 == 0 && x.first_thunk.0 == 0 { break; } },
-                Err(e) => return Err(e),
-            }
-
-            imports += 1;
-            address.0 += mem::size_of::<ImageImportDescriptor>() as u32;
-        }
-
-        Ok(imports)
-    }
-    
-    /// Get the import table pointed to by the [ImageDataDirectory](ImageDataDirectory) reference. This is typically used by the
-    /// [ImageDataDirectory::resolve](ImageDataDirectory::resolve) function.
-    pub fn parse_import_table<'data>(pe: &'data PE, dir: &'data ImageDataDirectory) -> Result<&'data [ImageImportDescriptor], Error> {
-        let size = match ImageImportDescriptor::parse_import_table_size(pe, dir) {
-            Ok(s) => s,
-            Err(e) => return Err(e),
-        };
-
-        let offset = match pe.translate(PETranslation::Memory(dir.virtual_address)) {
-            Ok(o) => o,
-            Err(e) => return Err(e),
-        };
-
-        pe.buffer.get_slice_ref::<ImageImportDescriptor>(offset, size)
-    }
-
-    /// Get the mutable import table pointed to by the [ImageDataDirectory](ImageDataDirectory) reference. This is typically used by the
-    /// [ImageDataDirectory::resolve_mut](ImageDataDirectory::resolve_mut) function.
-    pub fn parse_mut_import_table<'data>(pe: &'data mut PE, dir: &'data ImageDataDirectory) -> Result<&'data mut [ImageImportDescriptor], Error> {
-        let size = match ImageImportDescriptor::parse_import_table_size(pe, dir) {
-            Ok(s) => s,
-            Err(e) => return Err(e),
-        };
-
-        let offset = match pe.translate(PETranslation::Memory(dir.virtual_address)) {
-            Ok(o) => o,
-            Err(e) => return Err(e),
-        };
-
-        pe.buffer.get_mut_slice_ref::<ImageImportDescriptor>(offset, size)
-    }
-
     /// Get the thunk array pointed to by the ```original_first_thunk``` field.
     pub fn get_original_first_thunk<'data>(&self, pe: &'data PE) -> Result<Vec<Thunk<'data>>, Error> {
         self.parse_thunk_array(pe, self.original_first_thunk)
@@ -1321,7 +1233,7 @@ impl ImageImportDescriptor {
     }
 
     /// Get the imports represented by this import descriptor. This resolves the import table and returns a series of strings
-    /// representing both [ImageImportByName](ImageImportByName) structures as well as import ordinals.
+    /// representing both [`ImageImportByName`](ImageImportByName) structures as well as import ordinals.
     pub fn get_imports(&self, pe: &PE) -> Result<Vec<String>, Error> {
         let mut results = Vec::<String>::new();
 
@@ -1484,3 +1396,271 @@ impl Clone for ImageBaseRelocation {
         self.size_of_block = source.size_of_block;
     }
 }
+
+#[repr(packed)]
+pub struct ImageResourceDirectory {
+    pub characteristics: u32,
+    pub time_date_stamp: u32,
+    pub major_version: u16,
+    pub minor_version: u16,
+    pub number_of_named_entries: u16,
+    pub number_of_id_entries: u16,
+}
+impl ImageResourceDirectory {
+    /// Get the number of entries this resource directory has.
+    pub fn entries(&self) -> usize {
+        (self.number_of_named_entries as usize) + (self.number_of_id_entries as usize)
+    }
+}
+impl Clone for ImageResourceDirectory {
+    fn clone(&self) -> Self {
+        Self {
+            characteristics: self.characteristics,
+            time_date_stamp: self.time_date_stamp,
+            major_version: self.major_version,
+            minor_version: self.minor_version,
+            number_of_named_entries: self.number_of_named_entries,
+            number_of_id_entries: self.number_of_id_entries,
+        }
+    }
+    fn clone_from(&mut self, source: &Self) {
+        self.characteristics = source.characteristics;
+        self.time_date_stamp = source.time_date_stamp;
+        self.major_version = source.major_version;
+        self.minor_version = source.minor_version;
+        self.number_of_named_entries = source.number_of_named_entries;
+        self.number_of_id_entries = source.number_of_id_entries;
+    }
+}
+
+#[repr(packed)]
+pub struct ImageResourceDirectoryEntry {
+    pub name: FlaggedDword,
+    pub offset_to_data: FlaggedDword,
+}
+impl ImageResourceDirectoryEntry {
+    /// Get the ID of this directory entry.
+    ///
+    /// The ID can either be a [name](ResourceDirectoryID::Name) or a numeric [ID](ResourceDirectoryID::ID).
+    pub fn get_id(&self) -> ResourceDirectoryID {
+        if self.name.get_flag() {
+            ResourceDirectoryID::Name(ResourceOffset(self.name.get_dword()))
+        }
+        else {
+            ResourceDirectoryID::ID(self.name.get_dword())
+        }
+    }
+    /// Get the offset to the data this entry represents.
+    ///
+    /// This can be either [data](ResourceDirectoryData::Data) or [another directory](ResourceDirectoryData::Directory).
+    pub fn get_data(&self) -> ResourceDirectoryData {
+        if self.offset_to_data.get_flag() {
+            ResourceDirectoryData::Directory(ResourceOffset(self.offset_to_data.get_dword()))
+        }
+        else {
+            ResourceDirectoryData::Data(ResourceOffset(self.offset_to_data.get_dword()))
+        }
+    }
+}
+impl Clone for ImageResourceDirectoryEntry {
+    fn clone(&self) -> Self {
+        Self {
+            name: self.name,
+            offset_to_data: self.offset_to_data,
+        }
+    }
+    fn clone_from(&mut self, source: &Self) {
+        self.name = source.name;
+        self.offset_to_data = source.offset_to_data;
+    }
+}
+
+/// Represents an ```IMAGE_RESOURCE_DIR_STRING``` structure.
+///
+/// See [`ImageImportByName`](ImageImportByName) for an explanation as to why this structure
+/// is different from the others.
+#[derive(Clone, Debug)]
+pub struct ImageResourceDirString<'data> {
+    pub length: &'data u16,
+    pub name: &'data [CChar],
+}
+impl<'data> ImageResourceDirString<'data> {
+    /// Get a ```ImageResourceDirString``` object at the given RVA.
+    pub fn parse(pe: &'data PE, rva: RVA) -> Result<Self, Error> {
+        let offset = match pe.translate(PETranslation::Memory(rva)) {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+        
+        let length = match pe.buffer.get_ref::<u16>(offset) {
+            Ok(h) => h,
+            Err(e) => return Err(e),
+        };
+        
+        let name = match pe.buffer.get_slice_ref::<CChar>(Offset(offset.0 + (mem::size_of::<u16>() as u32)), *length as usize) {
+            Ok(n) => n,
+            Err(e) => return Err(e),
+        };
+
+        Ok(Self { length, name })
+    }
+}
+
+/// Represents a mutable ```IMAGE_RESOURCE_DIR_STRING``` structure.
+///
+/// See [`ImageImportByName`](ImageImportByName) for an explanation as to why this structure
+/// is different from the others.
+pub struct ImageResourceDirStringMut<'data> {
+    pub length: &'data mut u16,
+    pub name: &'data mut [CChar],
+}
+impl<'data> ImageResourceDirStringMut<'data> {
+    /// Get a mutable ```ImageResourceDirString``` object at the given RVA.
+    pub fn parse(pe: &'data mut PE, rva: RVA) -> Result<Self, Error> {
+        let mut offset = match pe.translate(PETranslation::Memory(rva)) {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+
+        unsafe {
+            let mut ptr = pe.buffer.offset_to_mut_ptr(offset);
+
+            let length = &mut *(ptr as *mut u16);
+            let u16_size = mem::size_of::<u16>();
+
+            ptr = ptr.add(u16_size);
+
+            if !pe.buffer.validate_ptr(ptr) {
+                return Err(Error::BadPointer);
+            }
+
+            offset.0 += u16_size as u32;
+            
+            let name = slice::from_raw_parts_mut(ptr as *mut CChar, *length as usize);
+
+            Ok(Self { length, name })
+        }
+    }
+}
+
+/// Represents an ```IMAGE_RESOURCE_DIR_STRING_U``` structure.
+///
+/// See [`ImageImportByName`](ImageImportByName) for an explanation as to why this structure
+/// is different from the others.
+#[derive(Clone, Debug)]
+pub struct ImageResourceDirStringU<'data> {
+    pub length: &'data u16,
+    pub name: &'data [WChar],
+}
+impl<'data> ImageResourceDirStringU<'data> {
+    /// Get a ```ImageResourceDirStringU``` object at the given RVA.
+    pub fn parse(pe: &'data PE, rva: RVA) -> Result<Self, Error> {
+        let offset = match pe.translate(PETranslation::Memory(rva)) {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+        
+        let length = match pe.buffer.get_ref::<u16>(offset) {
+            Ok(h) => h,
+            Err(e) => return Err(e),
+        };
+        
+        let name = match pe.buffer.get_slice_ref::<WChar>(Offset(offset.0 + (mem::size_of::<u16>() as u32)), *length as usize) {
+            Ok(n) => n,
+            Err(e) => return Err(e),
+        };
+
+        Ok(Self { length, name })
+    }
+}
+
+/// Represents a mutable ```IMAGE_RESOURCE_DIR_STRING_U``` structure.
+///
+/// See [`ImageImportByName`](ImageImportByName) for an explanation as to why this structure
+/// is different from the others.
+pub struct ImageResourceDirStringUMut<'data> {
+    pub length: &'data mut u16,
+    pub name: &'data mut [WChar],
+}
+impl<'data> ImageResourceDirStringUMut<'data> {
+    /// Get a mutable ```ImageResourceDirStringU``` object at the given RVA.
+    pub fn parse(pe: &'data mut PE, rva: RVA) -> Result<Self, Error> {
+        let mut offset = match pe.translate(PETranslation::Memory(rva)) {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+
+        unsafe {
+            let mut ptr = pe.buffer.offset_to_mut_ptr(offset);
+
+            let length = &mut *(ptr as *mut u16);
+            let u16_size = mem::size_of::<u16>();
+
+            ptr = ptr.add(u16_size);
+
+            if !pe.buffer.validate_ptr(ptr) {
+                return Err(Error::BadPointer);
+            }
+
+            offset.0 += u16_size as u32;
+            
+            let name = slice::from_raw_parts_mut(ptr as *mut WChar, *length as usize);
+
+            Ok(Self { length, name })
+        }
+    }
+}
+
+#[repr(packed)]
+pub struct ImageResourceDataEntry {
+    pub offset_to_data: RVA,
+    pub size: u32,
+    pub code_page: u32,
+    pub reserved: u32,
+}
+impl ImageResourceDataEntry {
+    /// Read the data pointed to by this data entry.
+    pub fn read<'data>(&self, pe: &'data PE) -> Result<&'data [u8], Error> {
+        if self.offset_to_data.0 == 0 || !pe.validate_rva(self.offset_to_data) {
+            return Err(Error::InvalidRVA);
+        }
+
+        let offset = match pe.translate(PETranslation::Memory(self.offset_to_data)) {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+
+        pe.buffer.read(offset, self.size as usize)
+    }
+    /// Read mutable data pointed to by this directory entry.
+    pub fn read_mut<'data>(&self, pe: &'data mut PE) -> Result<&'data mut [u8], Error> {
+        if self.offset_to_data.0 == 0 || !pe.validate_rva(self.offset_to_data) {
+            return Err(Error::InvalidRVA);
+        }
+
+        let offset = match pe.translate(PETranslation::Memory(self.offset_to_data)) {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+
+        pe.buffer.read_mut(offset, self.size as usize)
+    }
+    /// Write data to the directory entry. Returns [`Error::BufferTooSmall`](Error::BufferTooSmall) if the data
+    /// overflows the buffer provided by the directory entry.
+    pub fn write(&self, pe: &mut PE, data: &[u8]) -> Result<(), Error> {
+        if self.offset_to_data.0 == 0 || !pe.validate_rva(self.offset_to_data) {
+            return Err(Error::InvalidRVA);
+        }
+
+        let offset = match pe.translate(PETranslation::Memory(self.offset_to_data)) {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+
+        if (data.len() as u32) > self.size {
+            return Err(Error::BufferTooSmall);
+        }
+
+        pe.buffer.write(offset, data)
+    }
+}
